#!/usr/bin/env python3

"""
query robot state and publish position, velocity and effort values to /joint_states
"""

import rospy
from sensor_msgs.msg import JointState

class joinStatePub:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # get joints names and store them in dictionary
        self.joint_index_name_dic = kargs['rev_joints']
        # register this node in the network as a publisher in /joint_states topic
        self.pub_joint_states = rospy.Publisher('joint_states', JointState, queue_size=1)

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        # setup msg placeholder
        joint_msg = JointState()
        # get joint states
        for joint_index in self.joint_index_name_dic:
            # get joint state from pybullet
            joint_state = self.pb.getJointState(self.robot, joint_index)
            # fill msg
            joint_msg.name.append(self.joint_index_name_dic[joint_index])
            joint_msg.position.append(joint_state[0])
            joint_msg.velocity.append(joint_state[1])
            joint_msg.effort.append(joint_state[3]) # applied effort in last sim step
        # update msg time using ROS time api
        joint_msg.header.stamp = rospy.Time.now()
        # publish joint states to ROS
        self.pub_joint_states.publish(joint_msg)



import rospy
from geometry_msgs.msg import Twist


class DiffDrive:
    def __init__(self, pybullet, robot, **kwargs):
        self.pb = pybullet
        self.robot = robot

        try:
            name = kwargs["name"]
            parameters = rospy.get_param("~" + name)
            # parameters = {'plugins': [{'module': 'haibot_rosgymbullet.resources.plugins.body_vel_control', 'class': 'cmdVelCtrl'}, {'module': 'haibot_rosgymbullet.resources.plugins.odometry', 'class': 'simpleOdometry'}, {'module': 'haibot_rosgymbullet.resources.plugins.control', 'class': 'Control'}, {'module': 'haibot_rosgymbullet.resources.plugins.joint_state_pub', 'class': 'joinStatePub'}, {'module': 'haibot_rosgymbullet.resources.plugins.laser_scanner', 'class': 'laserScanner'}, {'module': 'haibot_rosgymbullet.resources.plugins.rgbd_camera', 'class': 'RGBDCamera'}, {'module': 'haibot_rosgymbullet.resources.plugins.diff_drive', 'class': 'DiffDrive', 'name': 'diff_drive_controller'}], 'loop_rate': 80.0, 'gravity': -9.81, 'max_effort': 10.0, 'use_intertia_from_file': False, 'laser': {'frame_id': 'base_scan', 'angle_min': -2.26889, 'angle_max': 2.26889, 'num_beams': 75, 'range_min': 0.05, 'range_max': 5.6, 'beam_visualisation': True}, 'rgbd_camera': {'frame_id': 'kinect_link', 'resolution': {'width': 640, 'height': 480}}, 'diff_drive_controller': {'left_joints': ['joint_left_wheel'], 'right_joints': ['joint_right_wheel'], 'wheel_separation': 1.0, 'wheel_radius': 0.05, 'cmd_vel': '/cmd_vel'}} + name
        except Exception:
            rospy.logerr("Not loading plugin {}".format(self.__class__.__name__))
            import traceback
            traceback.print_exc()
            self.no_execution = True
            return

        # joint information preparation
        self.revolute_joints_id_name = kwargs["rev_joints"]
        self.revolute_joints_name_id = {v: k for k, v in self.revolute_joints_id_name.items()}

        # a flag to monitor if anything went wrong
        self.no_execution = False

        cmd_vel_topic = parameters.get("cmd_vel", "~cmd_vel")
        self.cmd_vel_subscriber = rospy.Subscriber(cmd_vel_topic, Twist, self.cmd_vel_callback)

        if (wheel_separation := parameters.get("wheel_separation", None)) is None:
            rospy.logwarn("No wheel_separation provided, using 1.0 as default")
            wheel_separation = 1.0
        self.wheel_separation = wheel_separation
        if (wheel_radius := parameters.get("wheel_radius", None)) is None:
            rospy.logwarn("No wheel_radius provided, using 0.05 as default")
            wheel_radius = 0.05
        self.wheel_radius = wheel_radius

        if (left_joints := parameters.get("left_joints", None)) is None:
            rospy.logerr("No left_joints provided")
            self.no_execution = True
            return
        if isinstance(left_joints, list) is False and type(left_joints) == str:
            left_joints = [left_joints]
        if (right_joints := parameters.get("right_joints", None)) is None:
            rospy.logerr("No right_joints provided")
            self.no_execution = True
            return
        if isinstance(right_joints, list) is False and type(right_joints) == str:
            right_joints = [right_joints]

        self.joint_indices = {"left": [], "right": []}
        for side in ["left", "right"]:
            for joint_name in eval(side + "_joints"):
                if joint_name not in self.revolute_joints_name_id:
                    rospy.logerr("Could not find joint {} in urdf".format(joint_name))
                    self.no_execution = True
                    return
                self.joint_indices[side].append(self.revolute_joints_name_id[joint_name])

        self.cmd_vel: Twist = Twist()

    def cmd_vel_callback(self, msg: Twist):
        self.cmd_vel = msg

    def get_wheel_speeds(self):
        vx = self.cmd_vel.linear.x
        wz = self.cmd_vel.angular.z
        left = (vx - wz * self.wheel_separation / 2.0) / self.wheel_radius
        right = (vx + wz * self.wheel_separation / 2.0) / self.wheel_radius
        return {"left": left, "right": right}

    def execute(self):
        if not self.no_execution:
            speeds = self.get_wheel_speeds()
            for side in ["left", "right"]:
                indices = self.joint_indices[side]
                self.pb.setJointMotorControlArray(
                    bodyUniqueId=self.robot,
                    jointIndices=indices,
                    controlMode=self.pb.VELOCITY_CONTROL,
                    targetVelocities=[speeds[side] for _ in range(len(indices))]
                )


#!/usr/bin/env python3

"""
Laser scanner simulation based on pybullet rayTestBatch function
This code does not add noise to the laser scanner readings
"""

import rospy
import math
import numpy as np
from sensor_msgs.msg import LaserScan

class laserScanner:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # laser params
        laser_frame_id = 'base_scan' # laser reference frame, has to be an existing link
        # get pybullet laser link id from its name
        link_names_to_ids_dic = kargs['link_ids']
        if not laser_frame_id in link_names_to_ids_dic:
            rospy.logerr('laser reference frame "{}" not found in URDF model, cannot continue'.format(laser_frame_id))
            rospy.logwarn('Available frames are: {}'.format(link_names_to_ids_dic))
            rospy.signal_shutdown('required param frame id not set properly')
            return
        self.pb_laser_link_id = link_names_to_ids_dic[laser_frame_id]
        # create laser msg placeholder for publication
        self.laser_msg = LaserScan()
        # laser field of view
        # angle_min = rospy.get_param('~laser/angle_min', -1.5707963)
        angle_min = -2.26889
        # angle_max = rospy.get_param('~laser/angle_max', 1.5707963)
        angle_max = 2.26889
        assert(angle_max > angle_min)
        self.numRays = 640 # beams number
        self.laser_msg.range_min = rospy.get_param('~laser/range_min', 0.05)
        self.laser_msg.range_max = rospy.get_param('~laser/range_max', 10.0)
        self.beam_visualisation = True
        self.laser_msg.angle_min = angle_min
        self.laser_msg.angle_max = angle_max
        self.laser_msg.angle_increment = (angle_max - angle_min) / self.numRays
        # register this node in the network as a publisher in /scan topic
        self.pub_laser_scanner = rospy.Publisher('scan', LaserScan, queue_size=1)
        self.laser_msg.header.frame_id = laser_frame_id
        self.laser_msg.time_increment = 0.01 # ?
        self.laser_msg.scan_time = 0.01 # 10 hz
        # fixed_joint_index_name_dic = kargs['fixed_joints']
        self.rayHitColor = [0.7, 0.5, 0.7] # red color
        self.rayMissColor = [0, 1, 0] # green color
        # compute rays end beam position
        self.rayFrom, self.rayTo = self.prepare_rays()
        # variable used to run this plugin at a lower frequency, HACK
        self.count = 0

    def prepare_rays(self):
        """assume laser is in the origin and compute its x, y beam end position"""
        # prepare raycast origin and end values
        rayFrom = []
        rayTo = []
        for n in range(0, self.numRays):
            alpha = self.laser_msg.angle_min + n * self.laser_msg.angle_increment
            rayFrom.append([self.laser_msg.range_min * math.cos(alpha),
                          self.laser_msg.range_min * math.sin(alpha), 0.0])
            rayTo.append([self.laser_msg.range_max * math.cos(alpha),
                          self.laser_msg.range_max * math.sin(alpha), 0.0])
        return rayFrom, rayTo

    def transform_rays(self, laser_position, laser_orientation):
        """transform rays from reference frame using pybullet functions (not tf)"""
        laser_position = [laser_position[0], laser_position[1], laser_position[2]]
        TFrayFrom = []
        TFrayTo = []
        rm = self.pb.getMatrixFromQuaternion(laser_orientation)
        rotation_matrix = [[rm[0], rm[1], rm[2]],[rm[3], rm[4], rm[5]],[rm[6], rm[7], rm[8]]]
        for ray in self.rayFrom:
            position = np.dot(rotation_matrix, [ray[0], ray[1], ray[2]]) + laser_position
            TFrayFrom.append([position[0], position[1], position[2]])
        for ray in self.rayTo:
            position = np.dot(rotation_matrix, [ray[0], ray[1], ray[2]]) + laser_position
            TFrayTo.append([position[0], position[1], position[2]])
        return TFrayFrom, TFrayTo

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        # run at lower frequency, laser computations are expensive
        # self.count += 1
        # if self.count < 100:
        #     return
        # self.count = 0 # reset count

        # self.count += 0.1
        # if self.count < 10:
        #     return
        # self.count = 0.0 # reset count
        # remove any previous laser data if any
        self.laser_msg.ranges = []
        # remove previous beam lines from screen
        if self.beam_visualisation:
            self.pb.removeAllUserDebugItems()
        # get laser link position
        laser_state = self.pb.getLinkState(self.robot, self.pb_laser_link_id)
        # transform start and end position of the rays which were generated considering laser at the origin
        rayFrom, rayTo  = self.transform_rays(laser_state[0], laser_state[1]) # position + orientation
        # raycast using 4 threads
        results = self.pb.rayTestBatch(rayFrom, rayTo, 4)
        for i in range(self.numRays):
            if self.beam_visualisation:
                hitObjectUid = results[i][0]
                if hitObjectUid < 0:
                    # draw a line on pybullet gui for debug purposes in green because it did not hit any obstacle
                    self.pb.addUserDebugLine(rayFrom[i], rayTo[i], self.rayMissColor)
                else:
                    # draw a line on pybullet gui for debug purposes in red because it hited obstacle, results[i][3] -> hitPosition
                    self.pb.addUserDebugLine(rayFrom[i], results[i][3], self.rayHitColor)
            # compute laser ranges from hitFraction -> results[i][2]
            self.laser_msg.ranges.append(results[i][2] * self.laser_msg.range_max)
        # update laser time stamp with current time
        self.laser_msg.header.stamp = rospy.Time.now()
        # publish scan
        self.pub_laser_scanner.publish(self.laser_msg)



#!/usr/bin/env python3

"""
position, velocity and effort control for all revolute joints on the robot
"""

import rospy
from std_msgs.msg import Float64

# NOTE: 2 classes are implemented here, scroll down to the next class (Control) to see the plugin!

class pveControl:
    """helper class to receive position, velocity or effort (pve) control commands"""
    def __init__(self, joint_index, joint_name, controller_type):
        """constructor
        Assumes joint_name is unique, creates multiple subscribers to receive commands
        joint_index - stores an integer joint identifier
        joint_name - string with the name of the joint as described in urdf model
        controller_type - position, velocity or effort
        """
        assert(controller_type in ['position', 'velocity', 'effort'])
        rospy.Subscriber(joint_name + '_' + controller_type + '_controller/command',
                         Float64, self.pve_controlCB, queue_size=1)
        self.cmd = 0.0
        self.data_available = False
        self.joint_index = joint_index
        self.joint_name = joint_name

    def pve_controlCB(self, msg):
        """position, velocity or effort callback
        msg - the msg passed by the ROS network via topic publication
        """
        self.data_available = True
        self.cmd = msg.data

    def get_last_cmd(self):
        """method to fetch the last received command"""
        self.data_available = False
        return self.cmd

    def get_is_data_available(self):
        """method to retrieve flag to indicate that a command has been received"""
        return self.data_available

    def get_joint_name(self):
        """Unused method provided for completeness (pybullet works based upon joint index, not names)"""
        return self.joint_name

    def get_joint_index(self):
        """method used to retrieve the joint int index that this class points to"""
        return self.joint_index

# plugin is implemented below
class Control:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # lists to recall last received command (useful when controlling multiple joints)
        self.position_joint_commands = []
        self.velocity_joint_commands = []
        self.effort_joint_commands = []
        # this parameter will be set for all robot joints
        if rospy.has_param('~max_effort_vel_mode'):
            rospy.logwarn('max_effort_vel_mode parameter is deprecated, please use max_effort instead')
            # kept for backwards compatibility, delete after some time
            max_effort = rospy.get_param('~max_effort_vel_mode', 100.0)
        else:
            max_effort = rospy.get_param('~max_effort', 100.0)
        # the max force to apply to the joint, used in velocity control
        self.force_commands = []
        # get joints names and store them in dictionary, combine both revolute and prismatic dic
        self.joint_index_name_dic = {**kargs['rev_joints'], **kargs['prism_joints']}
        # setup subscribers
        self.pc_subscribers = []
        self.vc_subscribers = []
        self.ec_subscribers = []
        self.joint_indices = []
        # revolute joints - joint position, velocity and effort control command individual subscribers
        for joint_index in self.joint_index_name_dic:
            self.position_joint_commands.append(0.0)
            self.velocity_joint_commands.append(0.0)
            self.effort_joint_commands.append(0.0)
            # used only in velocity and position control mode
            self.force_commands.append(max_effort)
            # get joint name from dictionary, is used for naming the subscriber
            joint_name = self.joint_index_name_dic[joint_index]
            # create list of joints for later use in pve_ctrl_cmd(...)
            self.joint_indices.append(joint_index)
            # create position control object
            self.pc_subscribers.append(pveControl(joint_index, joint_name, 'position'))
            # create position control object
            self.vc_subscribers.append(pveControl(joint_index, joint_name, 'velocity'))
            # create position control object
            self.ec_subscribers.append(pveControl(joint_index, joint_name, 'effort'))

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        """check if user has commanded a joint and forward the request to pybullet"""
        # flag to indicate there are pending position control tasks
        position_ctrl_task = False
        velocity_ctrl_task = False
        effort_ctrl_task = False
        for index, subscriber in enumerate(self.pc_subscribers):
            if subscriber.get_is_data_available():
                self.position_joint_commands[index] = subscriber.get_last_cmd()
                position_ctrl_task = True
        for index, subscriber in enumerate(self.vc_subscribers):
            if subscriber.get_is_data_available():
                self.velocity_joint_commands[index] = subscriber.get_last_cmd()
                velocity_ctrl_task = True
        for index, subscriber in enumerate(self.ec_subscribers):
            if subscriber.get_is_data_available():
                self.effort_joint_commands[index] = subscriber.get_last_cmd()
                effort_ctrl_task = True
        # forward commands to pybullet, give priority to position control cmds, then vel, at last effort
        if position_ctrl_task:
            self.pb.setJointMotorControlArray(bodyUniqueId=self.robot, jointIndices=self.joint_indices,
                                     controlMode=self.pb.POSITION_CONTROL, targetPositions=self.position_joint_commands, forces=self.force_commands)
        elif velocity_ctrl_task:
            self.pb.setJointMotorControlArray(bodyUniqueId=self.robot, jointIndices=self.joint_indices,
                                     controlMode=self.pb.VELOCITY_CONTROL, targetVelocities=self.velocity_joint_commands, forces=self.force_commands)
        elif effort_ctrl_task:
            self.pb.setJointMotorControlArray(bodyUniqueId=self.robot, jointIndices=self.joint_indices,
                                     controlMode=self.pb.TORQUE_CONTROL, forces=self.effort_joint_commands)
        



#!/usr/bin/env python3

"""
Subscribe to cmd_vel and apply desired speed to the robot, without any noise

tf explained:
pybullet requires that velocity of the robot is set w.r.t. world reference frame
however cmd_vel convention required velocity to be expressed w.r.t. robot base frame
therefore a transformation is needed.
"""

import rospy
import math
import numpy as np

from geometry_msgs.msg import Twist, Vector3Stamped, Vector3

class cmdVelCtrl:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # subscribe to robot velocity commands
        self.cmd_vel_msg = None
        self.received_cmd_vel_time = None
        rospy.Subscriber("cmd_vel", Twist, self.cmdVelCB)

    # ---------- tf stuff starts

    def translation_matrix(self, direction):
        """copied from tf (listener.py)"""
        M = np.identity(4)
        M[:3, 3] = direction[:3]
        return M

    def quaternion_matrix(self, quaternion):
        """copied from tf (listener.py)"""
        epsilon = np.finfo(float).eps * 4.0
        q = np.array(quaternion[:4], dtype=np.float64, copy=True)
        nq = np.dot(q, q)
        if nq < epsilon:
            return np.identity(4)
        q *= math.sqrt(2.0 / nq)
        q = np.outer(q, q)
        return np.array((
            (1.0-q[1, 1]-q[2, 2],     q[0, 1]-q[2, 3],     q[0, 2]+q[1, 3], 0.0),
            (    q[0, 1]+q[2, 3], 1.0-q[0, 0]-q[2, 2],     q[1, 2]-q[0, 3], 0.0),
            (    q[0, 2]-q[1, 3],     q[1, 2]+q[0, 3], 1.0-q[0, 0]-q[1, 1], 0.0),
            (                0.0,                 0.0,                 0.0, 1.0)
            ), dtype=np.float64)

    def fromTranslationRotation(self, translation, rotation):
        """copied from tf (listener.py)"""
        return np.dot(self.translation_matrix(translation), self.quaternion_matrix(rotation))

    def lookupTransform(self, target_frame='base_link', source_frame='odom'):
        """
        copied from tf (listener.py)
        source_frame = odom, world_frame, target_frame = base_link, robot_frame
        """
        # get robot pose from pybullet
        t, r = self.pb.getBasePositionAndOrientation(self.robot)
        return [t[0], t[1], t[2]], [r[0], r[1], r[2], r[3]]

    def asMatrix(self, target_frame, hdr):
        """copied from tf (listener.py)"""
        translation,rotation = self.lookupTransform(target_frame, hdr.frame_id)
        return self.fromTranslationRotation(translation, rotation)

    def transformVector3(self, target_frame, v3s):
        """copied from tf (listener.py)"""
        mat44 = self.asMatrix(target_frame, v3s.header)
        mat44[0,3] = 0.0
        mat44[1,3] = 0.0
        mat44[2,3] = 0.0
        xyz = tuple(np.dot(mat44, np.array([v3s.vector.x, v3s.vector.y, v3s.vector.z, 1.0])))[:3]
        r = Vector3Stamped()
        r.header.stamp = v3s.header.stamp
        r.header.frame_id = target_frame
        r.vector = Vector3(*xyz)
        return r

    # ---------- tf stuff ends

    def cmd_vel_force_ctrler(self):
        """
        experimental code to set a torque on the robot model using a PID controller
        """
        pass
        #self.controller[0].setpoint = self.cmd_vel_msg.linear.x
        #self.controller[1].setpoint = self.cmd_vel_msg.linear.y
        #self.controller[2].setpoint = self.cmd_vel_msg.linear.z
        #self.controller[3].setpoint = self.cmd_vel_msg.angular.x
        #self.controller[4].setpoint = self.cmd_vel_msg.angular.y
        #self.controller[5].setpoint = self.cmd_vel_msg.angular.z
        ## query current robot speed from pybullet
        #linear_vel, angular_vel = self.pb.getBaseVelocity(self.robot)
        ## compute torque value based on PID library
        #output = []
        #for i in range(0, 3):
            #output.append(self.controller[i](linear_vel[i]))
        #for j in range(0, 3):
            #output.append(self.controller[j](angular_vel[j]))
        ## apply external force to robot body
        #self.z_offset = rospy.get_param('~cmd_vel_ctrl/z_offset', -0.35)
        #self.pb.applyExternalForce(self.robot, linkIndex=-1, forceObj=[output[0], output[1], output[2]],
                                    #posObj=[0.0, 0.0, self.z_offset], flags=self.pb.LINK_FRAME)
        #self.pb.applyExternalTorque(self.robot, linkIndex=-1, torqueObj=[output[3], output[4], output[5]],
                                    #flags=self.pb.LINK_FRAME)

    def cmdVelCB(self, msg):
        """callback to receive vel commands from user"""
        self.cmd_vel_msg = msg
        self.received_cmd_vel_time = rospy.Time.now()

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        if not self.cmd_vel_msg:
            return
        # check if timestamp is recent
        if (rospy.Time.now() - rospy.Duration(0.5)) > self.received_cmd_vel_time:
            return
        # transform Twist from base_link to odom (pybullet allows to set vel only on world ref frame)
        # NOTE: we would normally use tf for this, but there are issues currently between python 2 and 3 in ROS 1
        lin_vec = Vector3Stamped()
        lin_vec.header.frame_id = 'base_link'
        lin_vec.vector.x = self.cmd_vel_msg.linear.x
        lin_vec.vector.y = self.cmd_vel_msg.linear.y
        lin_vec.vector.z = self.cmd_vel_msg.linear.z
        ang_vec = Vector3Stamped()
        ang_vec.header.frame_id = 'base_link'
        ang_vec.vector.x = self.cmd_vel_msg.angular.x
        ang_vec.vector.y = self.cmd_vel_msg.angular.y
        ang_vec.vector.z = self.cmd_vel_msg.angular.z
        lin_cmd_vel_in_odom = self.transformVector3('odom', lin_vec)
        ang_cmd_vel_in_odom = self.transformVector3('odom', ang_vec)
        # set vel directly on robot model
        self.pb.resetBaseVelocity(self.robot, [lin_cmd_vel_in_odom.vector.x, lin_cmd_vel_in_odom.vector.y, lin_cmd_vel_in_odom.vector.z],
                                  [ang_cmd_vel_in_odom.vector.x, ang_cmd_vel_in_odom.vector.y, ang_cmd_vel_in_odom.vector.z])



#!usr/bin/env python3

import rospy
from geometry_msgs.msg import *
from sensor_msgs.msg import *
import gym
import importlib
from gym.utils import seeding
import numpy as np
import math
import time
import matplotlib.pyplot as plt
from std_srvs.srv import Empty
from haibot_rosgymbullet.resources.function_exec_manager import FuncExecManager
import os
import pybullet_data


class DiffBotDrivingEnv(gym.Env):
# class DiffBotDrivingEnv(object):
    metadata = {'render.modes': ['human']}

    def __init__(self):
        self.pb = importlib.import_module('pybullet')
        rospy.init_node('ros_gym_bullet', anonymous=False)
        self.plugins = []
        self.vel_pub = rospy.Publisher('/velocity', Twist, queue_size=5)
        #----------Action Space------------------
        # self.action_space = gym.spaces.box.Box(
        #     low=np.array([0, -.6], dtype=np.float32),
        #     high=np.array([1, .6], dtype=np.float32))
        # self.np_random, _ = seeding.np_random()

        # self.reward_range = (-np.inf, np.inf)

        # self.client =self.pb.connect(self.pb.DIRECT)
        # Reduce length of episodes for RL algorithms
        # self.pb.setTimeStep(1/30, self.client)

        self.loop_rate = rospy.get_param('~loop_rate', 80.0)
        self.pause_simulation = False
        rospy.Service('reset_simulation', Empty, self.handle_reset_simulation)
        # setup services for pausing/unpausing simulation
        rospy.Service('pause_physics', Empty, self.handle_pause_physics)
        rospy.Service('unpause_physics', Empty, self.handle_unpause_physics)
        # self.vel_pub = rospy.Publisher('cmd_vel', Twist, queue_size=5)
        is_gui_needed = True
        physicsClient = self.start_gui(gui=is_gui_needed)
        self.goal = None
        self.done = False
        # self.prev_dist_to_goal = None
        # self.rendered_img = None
        # self.render_rot_matrix = None
        # self.reset()
        # self.render()
        #------------------------------------------------------------------------------------

        # get pybullet path in your system and store it internally for future use, e.g. to set floor
        self.pb.setAdditionalSearchPath(pybullet_data.getDataPath())
        # create object of environment class for later use
        env_plugin = 'my_environment' # default : plugins/my_environment.py
        plugin_import_prefix = 'haibot_rosgymbullet.resources.plugins'
        self.environment = getattr(importlib.import_module(f'{plugin_import_prefix}.{env_plugin}'), 'Environment')(self.pb)
        # load robot URDF model, set gravity, and ground plane
        self.robot = self.init_pybullet_robot()
        self.connected_to_physics_server = None
        if not self.robot:
            self.connected_to_physics_server = False
            return # Error while loading urdf file
        else:
            self.connected_to_physics_server = True
        rev_joint_index_name_dic, prismatic_joint_index_name_dic, fixed_joint_index_name_dic, link_names_to_ids_dic = self.get_properties()
        # import plugins dynamically
        
        plugins = [ {'module': 'haibot_rosgymbullet.resources.plugins.body_vel_control', 'class': 'cmdVelCtrl'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.odometry', 'class': 'simpleOdometry'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.control', 'class': 'Control'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.joint_state_pub', 'class': 'joinStatePub'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.laser_scanner', 'class': 'laserScanner'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.rgbd_camera', 'class': 'RGBDCamera'},\
                    {'module': 'haibot_rosgymbullet.resources.plugins.diff_drive', 'class': 'DiffDrive', 'name': 'diff_drive_controller'} ]
        ########################################################################################################################################################################################
        # params = {'plugins': [{'module': 'haibot_rosgymbullet.resources.plugins.body_vel_control', 'class': 'cmdVelCtrl'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.odometry', 'class': 'simpleOdometry'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.control', 'class': 'Control'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.joint_state_pub', 'class': 'joinStatePub'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.laser_scanner', 'class': 'laserScanner'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.rgbd_camera', 'class': 'RGBDCamera'},\
        #             {'module': 'haibot_rosgymbullet.resources.plugins.diff_drive', 'class': 'DiffDrive', 'name': 'diff_drive_controller'}],\
        #         'loop_rate': 80.0, 'gravity': -9.81, 'max_effort': 10.0, 'use_intertia_from_file': False,\
        #         'laser': {'frame_id': 'base_scan', 'angle_min': -2.26889, 'angle_max': 2.26889, 'num_beams': 75,\
        #         'range_min': 0.05, 'range_max': 5.6, 'beam_visualisation': True}, 'rgbd_camera': {'frame_id': 'kinect_link', 'resolution': {'width': 640, 'height': 480}},\
        #         'diff_drive_controller': {'left_joints': ['joint_left_wheel'], 'right_joints': ['joint_right_wheel'], 'wheel_separation': 1.0, 'wheel_radius': 0.05, 'cmd_vel': '/cmd_vel'}}
        ########################################################################################################################################################################################
        # plugins = rospy.get_param('~plugins', [])
        if not plugins:
            rospy.logwarn('No plugins found, forgot to set param ~plugins?')
        # return to normal shell color
        print('\033[0m')
        # load plugins
        for plugin in plugins:
            module_ = plugin.pop("module")
            class_ = plugin.pop("class")
            params_ = plugin.copy()
            rospy.loginfo('loading plugin: {} class from {}'.format(class_, module_))
            # create object of the imported file class
            obj = getattr(importlib.import_module(module_), class_)(self.pb, self.robot,
                          rev_joints=rev_joint_index_name_dic,
                          prism_joints=prismatic_joint_index_name_dic,
                          fixed_joints=fixed_joint_index_name_dic,
                          link_ids=link_names_to_ids_dic,
                          **params_)
            # store objects in member variable for future use
            self.plugins.append(obj)

#---------------------------------------------------------------------------------------
    def init_pybullet_robot(self):
        """load robot URDF model, set gravity, ground plane and environment"""
        # get from param server the path to the URDF robot model to load at startup
        # urdf_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources/urdf", "diff_robot.urdf")
        urdf_path = "/home/saun/ReinforcementLearning_Robot/src/differential_robot/urdf/diff_robot.urdf"
        if urdf_path == None:
            rospy.signal_shutdown('mandatory param robot_urdf_path not set, will exit now')
        # test urdf file existance
        if not os.path.isfile(urdf_path):
            rospy.logerr('param robot_urdf_path is set, but file does not exist : ' + urdf_path)
            rospy.signal_shutdown('required robot urdf file not found')
            return None

        if 'xacro' in urdf_path:
            robot_description = rospy.get_param('robot_description', None)
            robot_description = "$(find xacro)/xacro $(arg urdf_path)"
            if not robot_description:
                rospy.logerr('required robot_description param not set')
                return None
            # remove xacro from name
            urdf_path_without_xacro = urdf_path[0:urdf_path.find('.xacro')]+urdf_path[urdf_path.find('.xacro')+len('.xacro'):]
            rospy.loginfo('generating urdf model from xacro from robot_description param server under: {0}'.format(urdf_path_without_xacro))
            try:
                urdf_file = open(urdf_path_without_xacro,'w')
            except:
                rospy.logerr('Failed to create urdf file from xacro, cannot write into destination: {0}'.format(urdf_path_without_xacro))
                return None
            urdf_file.write(robot_description)
            urdf_file.close()
            urdf_path = urdf_path_without_xacro
        # get robot spawn pose from parameter server
        robot_pose_x = rospy.get_param('~robot_pose_x', -2.3)
        robot_pose_y = rospy.get_param('~robot_pose_y', -1.0)
        robot_pose_z = rospy.get_param('~robot_pose_z', 0.0)
        robot_pose_yaw = rospy.get_param('~robot_pose_yaw', 0.0)
        robot_spawn_orientation = self.pb.getQuaternionFromEuler([0.0, 0.0, robot_pose_yaw])
        fixed_base = rospy.get_param('~fixed_base', False)
        # load robot from URDF model
        # user decides if inertia is computed automatically by pybullet or custom
        if rospy.get_param('~use_intertia_from_file', False):
            # combining several boolean flags using "or" according to pybullet documentation
            urdf_flags = self.pb.URDF_USE_INERTIA_FROM_FILE | self.pb.URDF_USE_SELF_COLLISION
        else:
            urdf_flags = self.pb.URDF_USE_SELF_COLLISION
        # load environment
        rospy.loginfo('loading environment')
        self.environment.load_environment()
        # set no realtime simulation, NOTE: no need to stepSimulation if setRealTimeSimulation is set to 1
        self.pb.setRealTimeSimulation(0) # NOTE: does not currently work with effort controller, thats why is left as 0
        rospy.loginfo('loading urdf model: ' + urdf_path)
        # NOTE: self collision enabled by default
        return self.pb.loadURDF(urdf_path, basePosition=[robot_pose_x, robot_pose_y, robot_pose_z],
                                           baseOrientation=robot_spawn_orientation,
                                           useFixedBase=fixed_base, flags=urdf_flags)

    def get_properties(self):
        rev_joint_index_name_dic = {}
        fixed_joint_index_name_dic = {}
        prismatic_joint_index_name_dic = {}
        link_names_to_ids_dic = {}
        for joint_index in range(0, self.pb.getNumJoints(self.robot)):
            info = self.pb.getJointInfo(self.robot, joint_index)
            # build a dictionary of link names to ids
            link_names_to_ids_dic[info[12].decode('utf-8')] = joint_index
            # ensure we are dealing with a revolute joint
            if info[2] == self.pb.JOINT_REVOLUTE:
                # insert key, value in dictionary (joint index, joint name)
                rev_joint_index_name_dic[joint_index] = info[1].decode('utf-8') # info[1] refers to joint name
            elif info[2] == self.pb.JOINT_FIXED:
                # insert key, value in dictionary (joint index, joint name)
                fixed_joint_index_name_dic[joint_index] = info[1].decode('utf-8') # info[1] refers to joint name
            elif info[2] == self.pb.JOINT_PRISMATIC:
                prismatic_joint_index_name_dic[joint_index] = info[1].decode('utf-8') # info[1] refers to joint name
        return rev_joint_index_name_dic, prismatic_joint_index_name_dic, fixed_joint_index_name_dic, link_names_to_ids_dic

    def handle_reset_simulation(self, req):
        """Callback to handle the service offered by this node to reset the simulation"""
        rospy.loginfo('reseting simulation now')
        self.pb.resetSimulation()
        return Empty()

    def start_gui(self, gui=True):
        """start physics engine (client) with or without gui"""
        if(gui):
            # start simulation with gui
            rospy.loginfo('Running pybullet with gui')
            rospy.loginfo('-------------------------')
            gui_options = rospy.get_param('~gui_options', '') # e.g. to maximize screen: options="--width=2560 --height=1440"
            return self.pb.connect(self.pb.GUI, options=gui_options)
        else:
            # start simulation without gui (non-graphical version)
            rospy.loginfo('Running pybullet without gui')
            # hide console output from pybullet
            rospy.loginfo('-------------------------')
            return self.pb.connect(self.pb.DIRECT)

    def handle_reset_simulation(self, req):
        rospy.loginfo('reseting simulation now')
        # pause simulation to prevent reading joint values with an empty world
        self.pause_simulation = True
        # remove all objects from the world and reset the world to initial conditions
        self.pb.resetSimulation()
        # load URDF model again, set gravity and floor
        self.init_pybullet_robot()
        # resume simulation control cycle now that a new robot is in place
        self.pause_simulation = False
        return []

    def handle_pause_physics(self, req):
        rospy.loginfo('pausing simulation')
        self.pause_simulation = False
        return []

    def handle_unpause_physics(self, req):
        rospy.loginfo('unpausing simulation')
        self.pause_simulation = True
        return []

    def pause_simulation_function(self):
        return self.pause_simulation

    def start_pybullet_ros_wrapper_sequential(self):
        """
        This function is deprecated, we recommend the use of parallel plugin execution
        """
        rate = rospy.Rate(self.loop_rate)
        while not rospy.is_shutdown():
            if not self.pause_simulation:
                # run x plugins
                for task in self.plugins:
                    task.execute()
                # perform all the actions in a single forward dynamics simulation step such
                # as collision detection, constraint solving and integration
                self.pb.stepSimulation()
            rate.sleep()
        rospy.logwarn('killing node now...')
        # if node is killed, disconnect
        if self.connected_to_physics_server:
            self.pb.disconnect()

    def start_pybullet_ros_wrapper_parallel(self):
        """
        Execute plugins in parallel, however watch their execution time and warn if exceeds the deadline (loop rate)
        """
        # create object of our parallel execution manager
        exec_manager_obj = FuncExecManager(self.plugins, rospy.is_shutdown, self.pb.stepSimulation, self.pause_simulation_function,
                                    log_info=rospy.loginfo, log_warn=rospy.logwarn, log_debug=rospy.logdebug, function_name='plugin')
        # start parallel execution of all "execute" class methods in a synchronous way
        exec_manager_obj.start_synchronous_execution(loop_rate=self.loop_rate)
        # ctrl + c was pressed, exit
        rospy.logwarn('killing node now...')
        # if node is killed, disconnect
        if self.connected_to_physics_server:
            self.pb.disconnect()

    def start_pybullet_ros_wrapper(self):
        if rospy.get_param('~parallel_plugin_execution', True):
            self.start_pybullet_ros_wrapper_parallel()
        else:
            self.start_pybullet_ros_wrapper_sequential()

#---------------------------------------------------------------------------------------
    def calculate_observation(self, data):
        min_range = 0.2
        done = False
        for i, item in enumerate(data.ranges):
            if (min_range > data.ranges[i] > 0):
                done = True
        return data.ranges, done
    

    def step(self, action):
        rospy.wait_for_service('/unpause_physics')
        try:
            unpause = rospy.ServiceProxy("/unpause_physics", Empty)
            unpause()
        except (rospy.ServiceException) as e:
            print ("/unpause_physics service call failed")

        max_ang_speed = 0.3
        ang_vel = (action-10)*max_ang_speed*0.1 #from (-0.33 to + 0.33)

        vel_cmd = Twist()
        vel_cmd.linear.x = 0.2
        vel_cmd.angular.z = ang_vel
        self.vel_pub.publish(vel_cmd)

        data = None
        while data is None:
            try:
                data = rospy.wait_for_message('/scan', LaserScan, timeout=5)
            except:
                pass

        rospy.wait_for_service('/pause_physics')
        try:
            pause = rospy.ServiceProxy("/pause_physics", Empty)
            pause()
        except (rospy.ServiceException) as e:
            print ("/pause_physics service call failed")

        state, done = self.calculate_observation(data)

        if not done:
            # Straight reward = 5, Max angle reward = 0.5
            reward = round(15*(max_ang_speed - abs(ang_vel) +0.0335), 2)
            # print ("Action : "+str(action)+" Ang_vel : "+str(ang_vel)+" reward="+str(reward))
        else:
            reward = -200

        return np.asarray(state), reward, done, {}


    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]
    

    def reset(self):
        # Resets the state of the environment and returns an initial observation.
        rospy.wait_for_service('/reset_simulation')
        try:
            reset_proxy = rospy.ServiceProxy('/reset_simulation', Empty)
            reset_proxy()
        except (rospy.ServiceException) as e:
            print ("/reset_simulation service call failed")

        # Unpause simulation to make observation
        rospy.wait_for_service('/unpause_physics')
        try:
            unpause = rospy.ServiceProxy("/unpause_physics", Empty)
            unpause()
        except (rospy.ServiceException) as e:
            print ("/unpause_physics service call failed")
        #read laser data
        data = None
        while data is None:
            try:
                data = rospy.wait_for_message('/scan', LaserScan, timeout=5)
            except:
                pass

        rospy.wait_for_service('/pause_physics')
        try:
            pause = rospy.ServiceProxy("/pause_physics", Empty)
            pause()
        except (rospy.ServiceException) as e:
            print ("/pause_physics service call failed")

        state, done = self.calculate_observation(data)

        return np.asarray(state)

    def render(self, mode='human'):
        # if self.rendered_img is None:
        #     self.rendered_img = plt.imshow(np.zeros((100, 100, 4)))

        # Base information
        # DiffBot_id, client_id = self.robot.get_ids()
        # proj_matrix =self.pb.computeProjectionMatrixFOV(fov=80, aspect=1,
        #                                            nearVal=0.01, farVal=100)
        # pos, ori = [list(l) for l in
        #            self.pb.getBasePositionAndOrientation(DiffBot_id, client_id)]
        # pos[2] = 0.2
        # abc = DiffBotDrivingEnv()
        
        self.start_pybullet_ros_wrapper()

    def close(self):
       self.pb.disconnect(self.client)
